local Sackville = LibStub("AceAddon-3.0"):GetAddon("Sackville", "AceConsole-3.0")
local B = LibStub("LibBabble-Inventory-3.0")
local BL = B:GetLookupTable()
-- local PT = LibStub("LibPeriodicTable-3.1")

local sprintf = string.format

Sackville.Filter = {}
local allFilters = {}
local itemQualities = {
    [ITEM_QUALITY0_DESC] = 0,
    [ITEM_QUALITY1_DESC] = 1,
    [ITEM_QUALITY2_DESC] = 2,
    [ITEM_QUALITY3_DESC] = 3,
    [ITEM_QUALITY4_DESC] = 4,
    [ITEM_QUALITY5_DESC] = 5,
    [ITEM_QUALITY6_DESC] = 6,
    [ITEM_QUALITY7_DESC] = 7
}

local function AlwaysFalse()
    return false
end

local function IsArmor(item)
    return allFilters[BL["Armor"]].filter(item)
end

local function IsShirt(item)
    return item.equipLoc == "INVTYPE_BODY"
end

local function IsTrinket(item)
    return item.equipLoc == "INVTYPE_TRINKET"
end

local function IsNecklace(item)
    return (item.equipLoc == "INVTYPE_NECK")
end

local function IsRing(item)
    return (item.equipLoc == "INVTYPE_FINGER")
end

local function IsTabard(item)
    -- Sackville.Print("Comparing equipLoc " .. item.equipLoc .. " to INVTYPE_TABARD")
    return item.equipLoc == "INVTYPE_TABARD"
end

local function IsEquipment(item)
    if item.equipLoc == "" then
        return false
    else
        return true
    end
end

local function MakeQualityFilter(quality)
    local code = "return function (item) " .. string.format(      
        [[return item.quality == "%d" end]], quality)
    return assert(loadstring(code)())
end

local function MakeStandardFilter(category, subCategory)
    assert(category ~= nil)

    local code = "return function (item) "    
    if subCategory then
        code = code .. string.format([[
            return (item.type == "%s" and item.subType == "%s")]],
            category, subCategory)
    else
        code = code .. string.format([[
            return (item.type == "%s")]], category)
    end
    
    code = code .. " end"
    return assert(loadstring(code)())
end

function Sackville.Filter:AddFilter(category, catFilter, subCategory, subCatFilter)
-- category: top level category
-- catFilter: function returning true/false if the tested item is a member
-- subCategory: second level category
-- subCatFilter: function returning true/false if the tested item is a member
    if not allFilters[category] then
        allFilters[category] = {}
        allFilters[category].subCategories = {}
        allFilters[category].filter = catFilter
    end

    if subCategory then
        Sackville:Debug(2, "filter", "AddFilter cat=%s, subCat=%s",
            category, subCategory)
        allFilters[category].subCategories[subCategory] = {}
        allFilters[category].subCategories[subCategory].filter = subCatFilter
    end
end

function DeepCopy(t) -- not used
    if type(t) ~= 'table' then return t end
    local mt = getmetatable(t)
    local res = {}
    for k,v in pairs(t) do
        if type(v) == 'table' then
        v = deepcopy(v)
        end
        res[k] = v
    end
    setmetatable(res,mt)
    return res
end

function Sackville.Filter:ItemMatchesSection(item, section)
    if section.isDefault then
        Sackville:Debug(3, "filter", "* Hit default section: %s",section.name)
        return true
    end
    
    for _, catName in ipairs(section.categories) do
        local matchRules = Sackville.Options.categories[catName]
        assert(matchRules, "nil matchRules, catName:" .. catName)
        if matchRules and #matchRules > 0 then
            local isMatch = Sackville.Filter:ItemMatchesRules(item, matchRules)
            if isMatch then return true end
        else
            Sackville:Debug(3, "filter", "* Found no rules in section:", section.name)        
        end
    end
    
    return false
end

function Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex)
    ruleIndex = ruleIndex or 1
    
    local rule = matchRules[ruleIndex]
    local returnValue = false
    
    if rule then
        assert(rule.operation == "and" or rule.operation == "or")
        Sackville:Debug(3, "filter", "- Processing rule index " .. ruleIndex)
        if matchRules[ruleIndex+1] then
        
            if rule.operation == "and" then
                if rule.subCategory then
                    local objCat = allFilters[rule.category]
                    local objSubCat = objCat.subCategories[rule.subCategory]
                    
                    returnValue = objSubCat.filter(item) and
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    Sackville:Debug(3, "filter", "- Trying filter %s/%s: %s",
                        rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")
                else
                    local objCat = allFilters[rule.category]
                    returnValue = objCat.filter(item) and
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    Sackville:Debug(3, "filter", "- Trying filter %s/%s: %s",
                        rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")
                end
            else -- "or"
                if rule.subCategory then
                    local objCat = allFilters[rule.category]
                    local objSubCat = objCat.subCategories[rule.subCategory]
                    returnValue = objSubCat.filter(item) or
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    Sackville:Debug(3, "filter", "- Trying filter %s/%s: %s",
                        rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")
                else
                    local objCat = allFilters[rule.category]
                    returnValue = objCat.filter(item) or
                        Sackville.Filter:ItemMatchesRules(item, matchRules, ruleIndex+1)
                    Sackville:Debug(3, "filter", "- Trying filter %s/%s: %s",
                        rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")
                end            
            end
        else -- This is the last (and maybe only) rule
            if rule.subCategory then
                local objCat = allFilters[rule.category]
                local objSubCat = objCat.subCategories[rule.subCategory]
                returnValue = objSubCat.filter(item)
                Sackville:Debug(3, "filter", "- Trying filter %s/%s: %s",
                        rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")
            else
                local objCat = allFilters[rule.category]
                returnValue = objCat.filter(item)
                Sackville:Debug(3, "filter", "- Trying filter %s/%s: %s",
                        rule.category, rule.subCategory or "", returnValue and "TRUE" or "FALSE")
            end       
        end
        
        if rule.negate then
            Sackville:Debug(3, "filter", "- NEGATING RETURN VALUE")
            returnValue = not returnValue
        end
    end
    

    return returnValue
end

-- Populate allFilters with filters based on the primary and secondary
-- categories being equal to item.type and item.subType
for category, subCategories in pairs(Sackville.Options.standardCategories) do
    local catFilter = MakeStandardFilter(category)
    
    for _, subCategory in ipairs(subCategories) do
        local subCatFilter = MakeStandardFilter(category, subCategory)
        Sackville.Filter:AddFilter(category, catFilter, subCategory, subCatFilter)
    end
end

for name, value in pairs(itemQualities) do
    -- FIXME: Localize "Quality"
    Sackville.Filter:AddFilter("Quality", AlwaysFalse, name, MakeQualityFilter(value))
end

Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Armor"], IsArmor)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Trinket"], IsTrinket)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Shirt"], IsShirt)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Neck"], IsNecklace)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Ring"], IsRing)
Sackville.Filter:AddFilter("Equipment", IsEquipment, BL["Tabard"], IsTabard)